﻿using UnityEngine;
using System.Collections; 
using System;
using System.Runtime.InteropServices;
using System.IO;

public class trackerInstance : MonoBehaviour {

	[ StructLayout( LayoutKind.Sequential, Pack=0 )] // set Pack=0 for Windows and Pack=1 for OSX
	public struct TrackerReport
	{
		public VrpnManager.TimeVal msg_time;
		public int sensor;
		[ MarshalAs( UnmanagedType.ByValArray, SizeConst=3 )]
		public double[] pos;
		[ MarshalAs( UnmanagedType.ByValArray, SizeConst=4 )]
		public double[] quat;
	}
	
	// Class Properties
	public static int num_trackers = 0;
	public enum Derivation_Type { None=1, Velocity=2, Acceleration=4 };          
	public enum Transform_Type { None=1, Position=2, Orientation=3, Both=4 }; 
	public bool trackRelativeToObjectPos = false;

    //Object Properties
    private Vector3 objectInitialPos = Vector3.zero;
    private Quaternion objectInitialQuat = Quaternion.identity;

    //Tracker Properties
    private bool hasFoundFirstPos = false;

	// Public Properties
	public VrpnManager.Tracker_Types TrackerType;
	public string TrackerName;
	public Transform_Type ApplyTracking = Transform_Type.Both; 
	public Derivation_Type Derivation = Derivation_Type.None;  
	public Transform DeviceToUnity;
	public Transform SensorOffset;   
	public int SensorNumber = 0;
	public int MaxReports = 20;
	public bool ShowDebug = false;
	
	// Private Variables
	private bool initialized = false;
	private VrpnManager.TimeVal LastReport;
	private IntPtr[] repsPtr;
	
	private Vector3 lastPos = Vector3.zero;
	private Quaternion lastQuat = Quaternion.identity;
	private Vector3 trackerPos = Vector3.zero;
	private Quaternion trackerQuat = Quaternion.identity;
	private Vector3 pos = Vector3.zero;
	private Quaternion quat = Quaternion.identity;
    private Vector3 trackerFirstPos = Vector3.zero;
    private Quaternion trackerFirstQuat = Quaternion.identity;

    private Vector3 scalePos = Vector3.zero;
	
	private string debug_text = "";
	private int debug_xoffset;
	
	public Vector3 GetPosition()
	{
		return trackerPos;
	}
	
	public Quaternion GetRotation()
	{
		return trackerQuat;
	}

	[DllImport("E:/Users/CafeMedia/Google Drive/Skoleting/Masteroppgave/Unity/Assets/Plugins/VRPN/VrpnProject.dll")]
	private static extern int VRPNTrackerStart(string name, int deriv);
	
	void Start () {

        //Get the object position and rotation:
        objectInitialPos = transform.position;
        objectInitialQuat = transform.rotation;

        scalePos[0] = -1;
        scalePos[1] = 1;
        scalePos[2] = 1;

		//allocate unmanaged memory for tracker reports
		repsPtr = new IntPtr[MaxReports];
		TrackerReport report = new TrackerReport();
		report.sensor = 0;
		report.pos = new double[3];
		report.quat = new double[4];
		report.quat[3] = 1.0f;
		for(int i=0; i<MaxReports ; i++)
		{
			repsPtr[i] = Marshal.AllocHGlobal(Marshal.SizeOf(typeof(TrackerReport)));
			Marshal.StructureToPtr(report, repsPtr[i], true);
		}
		if (DeviceToUnity == null || SensorOffset == null)
		{
			if (DeviceToUnity == null)
				Console.WriteLine("Warning: DeviceToUnity field not referencing a transform");
			if (SensorOffset == null)
				Console.WriteLine("Warning: SensorOffset field not referencing a transform");
		}
		
		// Setup last report time memory 
		LastReport = new VrpnManager.TimeVal();  
	}
	
	bool StartTracker () { 	
		if (VrpnManager.initialized)
		{
			// Register Tracker Device
			int debug = VRPNTrackerStart(TrackerName,(int)Derivation);
			num_trackers++;
			initialized = true;
            //Console.WriteLine(debug.ToString());
			return true;
		}
		return false;
	}
	
	void Update () {
		// Ensure device is ready
		if (!initialized && !StartTracker()) return;

        //Debug.Log("TrackerInstance: Laawrd Almaighty");
		GetLatestPosQuat();
		//GetAveragePosQuat();

        if (!hasFoundFirstPos) //Find first position
        {
            trackerFirstPos = trackerPos;
            trackerFirstQuat = trackerQuat;
            if (ShowDebug)
            {
                debug_text = this.name + System.String.Format("\nORI[X:{0:F2},Y:{1:F2},Z:{2:F2}]", trackerFirstPos.x, trackerFirstPos.y, trackerFirstPos.z);
                //System.String.Format("\nPOS[X:{0:F2},Y:{1:F2},Z:{2:F2}]", pos.x, pos.y, pos.z) +
                Debug.Log(debug_text);
            }
            hasFoundFirstPos = true;
        }

		if (ApplyTracking == Transform_Type.Both)
		{
            pos = (trackerPos - SensorOffset.position);
            pos.x = pos.x * -1;
            lastPos = pos;

            //Interchange x and z orientation 
            quat = trackerQuat;
            quat.z = quat.z * -1;

            lastQuat = quat;
			UpdateTransform(pos, quat);
		}
		else if (ApplyTracking == Transform_Type.Position)
		{
            if (trackRelativeToObjectPos)
            {
                pos = ((trackerPos - trackerFirstPos) - SensorOffset.position);
                pos.x = pos.x * -1;
                lastPos = pos;
                quat = lastQuat;
            }
            else
            {
                pos = trackerPos;
                lastPos = pos;
                quat = lastQuat;
            }
			UpdateTransform(pos, quat);
		}
		else if (ApplyTracking == Transform_Type.Orientation)
		{
			pos = lastPos;
			quat = trackerQuat;
			lastQuat = quat;
			UpdateTransform(pos, quat);
		}
		else
		{
			//pos = lastPos;
			//quat = lastQuat;
		}

        

        /*if (ShowDebug)
        {
            debug_text = this.name + System.String.Format("\nORI[X:{0:F2},Y:{1:F2},Z:{2:F2}]", quat.eulerAngles[0], quat.eulerAngles[1], quat.eulerAngles[2]);
            //System.String.Format("\nPOS[X:{0:F2},Y:{1:F2},Z:{2:F2}]", pos.x, pos.y, pos.z) +
            Debug.Log(debug_text);
        }*/
	}

	protected virtual void UpdateTransform(Vector3 pos, Quaternion quat)
	{
		if (ApplyTracking == Transform_Type.Both || ApplyTracking == Transform_Type.Position) {
			transform.localPosition = pos;

		}
		if (ApplyTracking == Transform_Type.Both || ApplyTracking == Transform_Type.Orientation) {
			transform.localRotation = quat;
		}
	}

	[DllImport("E:/Users/CafeMedia/Google Drive/Skoleting/Masteroppgave/Unity/Assets/Plugins/VRPN/VrpnProject.dll")]
	private static extern int VRPNTrackerPosReport(string name, [In,Out] IntPtr rep, [Out] IntPtr ts, int sensor);
	
	void GetLatestPosQuat()
	{
        VRPNTrackerPosReport(TrackerName,repsPtr[0],IntPtr.Zero,SensorNumber);
		TrackerReport rep = (TrackerReport)Marshal.PtrToStructure(repsPtr[0],typeof(TrackerReport));
        trackerPos.x = (float)rep.pos[0];
		trackerPos.y = (float)rep.pos[1];
		trackerPos.z = (float)rep.pos[2];
		trackerQuat.x = (float)rep.quat[0];	//CHANGED THIS WITH THE Z ONE
		trackerQuat.y = -(float)rep.quat[1];	
		trackerQuat.z = (float)rep.quat[2];	
		trackerQuat.w = (float)rep.quat[3];
		LastReport.tv_sec = rep.msg_time.tv_sec;
		LastReport.tv_usec = rep.msg_time.tv_usec;
	}

	[DllImport("E:/Users/CafeMedia/Google Drive/Skoleting/Masteroppgave/Unity/Assets/Plugins/VRPN/VrpnProject.dll")]
	private static extern int VRPNTrackerNumPosReports(string name);

	[DllImport("E:/Users/CafeMedia/Google Drive/Skoleting/Masteroppgave/Unity/Assets/Plugins/VRPN/VrpnProject.dll")]
	private static extern void VRPNTrackerPosReports(string name, [In,Out] IntPtr[] repsPtr, [In,Out] ref int nmbr);
	
	void GetAveragePosQuat()
	{
		int num = MaxReports;
		VRPNTrackerPosReports(TrackerName,repsPtr,ref num);
		float[] repsSum = new float[7];
		int repsCount = 0;
		TrackerReport[] reps = new TrackerReport[num];
		for (int i=0; i<num; i++)
		{
			reps[i] = (TrackerReport)Marshal.PtrToStructure(repsPtr[i],typeof(TrackerReport));
			if (reps[i].sensor == SensorNumber && VrpnManager.TimeValGreater(ref reps[i].msg_time,ref LastReport))
			{
				repsSum[0] += (float)reps[i].pos[0];
				repsSum[1] += (float)reps[i].pos[1];
				repsSum[2] += (float)reps[i].pos[2];
				repsSum[3] += (float)reps[i].quat[2];	
				repsSum[4] += -(float)reps[i].quat[1];	
				repsSum[5] += (float)reps[i].quat[0];	
				repsSum[6] += (float)reps[i].quat[3];
				LastReport.tv_sec = reps[i].msg_time.tv_sec;
				LastReport.tv_usec = reps[i].msg_time.tv_usec;
				repsCount++;
			}
		}
		if (repsCount > 0)
		{
			trackerPos.x = repsSum[0]/(float)repsCount;
			trackerPos.y = repsSum[1]/(float)repsCount;
			trackerPos.z = repsSum[2]/(float)repsCount;
			trackerQuat.x = repsSum[3]/(float)repsCount;	
			trackerQuat.y = repsSum[4]/(float)repsCount;	
			trackerQuat.z = repsSum[5]/(float)repsCount;	
			trackerQuat.w = repsSum[6]/(float)repsCount;
		}
		else
		{
			trackerPos = lastPos;
			trackerQuat = lastQuat;
		}
	}
	
	Vector3 LatLongToMeter(float latitude) 
	{
		// Convert latitude to radians
		Vector3 result = Vector3.one;
		double lat = (latitude * 2.0 * Math.PI)/360.0;
		
		// Set up "Constants"
		double m1 = 111132.92;		// latitude calculation term 1
		double m2 = -559.82;		// latitude calculation term 2
		double m3 = 1.175;			// latitude calculation term 3
		double m4 = -0.0023;		// latitude calculation term 4
		double p1 = 111412.84;		// longitude calculation term 1
		double p2 = -93.5;			// longitude calculation term 2
		double p3 = 0.118;			// longitude calculation term 3
		
		// Calculate the length of a degree of latitude and longitude in meters
		double latlen = m1 + (m2 * Math.Cos(2 * lat)) + (m3 * Math.Cos(4 * lat)) + (m4 * Math.Cos(6 * lat));
		double longlen = (p1 * Math.Cos(lat)) + (p2 * Math.Cos(3 * lat)) + (p3 * Math.Cos(5 * lat));
		
		result.x = (float)longlen;
		result.z = (float)latlen;
		return result;
	}
	
	public static Quaternion QuaternionFromMatrix(Matrix4x4 m) 
	{
		// Adapted from: http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
		Quaternion q = new Quaternion();
		q.w = Mathf.Sqrt( Mathf.Max( 0, 1 + m[0,0] + m[1,1] + m[2,2] ) ) / 2;
		q.x = Mathf.Sqrt( Mathf.Max( 0, 1 + m[0,0] - m[1,1] - m[2,2] ) ) / 2;
		q.y = Mathf.Sqrt( Mathf.Max( 0, 1 - m[0,0] + m[1,1] - m[2,2] ) ) / 2;
		q.z = Mathf.Sqrt( Mathf.Max( 0, 1 - m[0,0] - m[1,1] + m[2,2] ) ) / 2;
		q.x *= Mathf.Sign( q.x * ( m[2,1] - m[1,2] ) );
		q.y *= Mathf.Sign( q.y * ( m[0,2] - m[2,0] ) );
		q.z *= Mathf.Sign( q.z * ( m[1,0] - m[0,1] ) );
		return q;
	}
	
}
